
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifndef MAX_PATH
# define MAX_PATH 260
#endif

/* ADPCM */
#define XA_DATA_START   (0x44-48)
#define XA_CHANNELS 256

typedef char XAGroup[128];

typedef struct XASector {
	char         sectorFiller[48];
	XAGroup      XAGroups[18];
} XASector;

int checkXaHeader(char *xa_hdr)
{
	if (memcmp(xa_hdr, "RIFF", 4))
		return 1;
	if (memcmp(&xa_hdr[8], "CDXA", 4))
		return 1;
	return 0;
}

FILE * fopen_wrap(const char *name, const char *type)
{
    FILE *dump = fopen(name, type);

    if (dump == NULL)
        fprintf(stderr, "Error opening \"%s\" as \"%s\": File missing or inaccessible.\n", name, type);

    return dump;
}

int main(const int argc, const char *argv[])
{
	char xa_hdr[XA_DATA_START];
    XASector ssct;

    XASector *sectors[XA_CHANNELS];
    int sectorcount[XA_CHANNELS];
    int samplerate[XA_CHANNELS];
    int channels[XA_CHANNELS];

    FILE *in;
    const char *xaName;

    if (argc < 2)
        return 1;

    memset(sectors, 0, sizeof(XASector*) * XA_CHANNELS);
    memset(&sectorcount, 0, sizeof(int) * XA_CHANNELS);
    memset(&samplerate, 0, sizeof(int) * XA_CHANNELS);
    memset(&channels, 0, sizeof(int) * XA_CHANNELS);

    xaName = argv[1];

    in = fopen_wrap(xaName, "rb");

    if (!in)
        return 2;

	if (fread(&xa_hdr, sizeof(xa_hdr), 1, in) != 1)
    {
		fprintf(stderr, "Error reading input file.\n");
		return 3;
	}
	if (checkXaHeader(xa_hdr))
    {
		fprintf(stderr, "Input file not XA format.\n");
		return 4;
	}

	memset(&ssct, 0, sizeof(XASector));
	while (fread(&ssct, sizeof(XASector), 1, in) == 1)
	{
#define SUBMODE_REAL_TIME_SECTOR (1 << 6)
#define SUBMODE_FORM             (1 << 5)
#define SUBMODE_AUDIO_DATA       (1 << 2)
		if (ssct.sectorFiller[46] != (SUBMODE_REAL_TIME_SECTOR | SUBMODE_FORM | SUBMODE_AUDIO_DATA)) // submode
            continue;

        const int channel = ssct.sectorFiller[45];

        sectors[channel] = (XASector*) realloc(sectors[channel], (sectorcount[channel] + 1) * sizeof(XASector));

        ssct.sectorFiller[45] = 0;

        memcpy (&sectors[channel][sectorcount[channel]], &ssct, sizeof(XASector));

        sectorcount[channel]++;
    }

    for (int i = 0; i < XA_CHANNELS; ++i)
    {
        if (sectorcount[i] > 0)
        {
            char namebuf[MAX_PATH];

            snprintf(namebuf, MAX_PATH, "%s.%03i.XA", xaName, i);

            FILE *out = fopen_wrap(namebuf, "wb");

            fwrite(&xa_hdr, sizeof(xa_hdr), 1, out);

            for (int j = 0; j < sectorcount[i]; ++j)
                fwrite(&sectors[i][j], sizeof(XASector), 1, out);

            fclose(out);
        }
    }

    return 0;
}

